#include <stdio.h>
#include "bldc.h"

typedef enum {
  /* Motor is stopped (w/ brake) */
  MOTOR_ESTOP,
  /* Motor is stopped (unpowered) */
  MOTOR_STOP,
  /* Quick travel to target location, then hold */
  MOTOR_TRAVEL,
  /* Track step/direction input */
  MOTOR_TRACK,
  /* Motor maintains a constant RPM */
  MOTOR_RPM,
  /* Motor maintains a constant TORQUE */
  MOTOR_TORQUE,
  /* used during calibration */
  MOTOR_CALIBRATE,
  /* used during rpm loop calibration */
  MOTOR_CALIBRATE_RPM_LOOP,
  /* used during current loop calibration */
  MOTOR_CALIBRATE_CURRENT_LOOP
} Motor_Mode;

static Motor_Mode motor_mode = MOTOR_CALIBRATE;
volatile Position target_position;
static RPM target_rpm;
static float target_torque;

volatile int motor_pwm_ticks;
volatile int motor_pwm_rpm;
volatile Position motor_pwm_position;

static Position rpm_origin_pos;
static long rpm_origin_msec;
static float rpm_pos_per_loop;
static float rpm_delta;
static int i_term = 0;

volatile float desired_rpm;
volatile float desired_current;
volatile float calib_desired_current;
volatile float calib_desired_rpm;

typedef struct {
  Pid_Data *pid;
  float i_val;
  int i_enabled;
} Pid_State;

Pid_State pos_pid;
Pid_State rpm_pid;
Pid_State q_pid;
Pid_State d_pid;

static void
pid_init (Pid_State *pid, Pid_Data *data)
{
  pid->pid = data;
  pid->i_val = 0;
  pid->i_enabled = 1;
}

static inline float
pid_iter (float desired, float actual, float ff_val, Pid_State *pid)
{
  float e = desired - actual;
  if (pid->i_enabled)
    {
      pid->i_val += e / 20000.0 * pid->pid->i;
      if (pid->i_val < -pid->pid->imax)
	pid->i_val = -pid->pid->imax;
      if (pid->i_val > pid->pid->imax)
	pid->i_val = pid->pid->imax;
    }
  return pid->pid->p * e + pid->i_val + pid->pid->ff * ff_val;
}

static inline void
pid_reset (Pid_State *pid)
{
  pid->i_val = 0;
  pid->i_enabled = 1;
}

State state[NSTATE+2];
int state_n;

#define SQRT_3 1.732050807

static void
clarke_transform (float u_c, float v_c, float w_c,
		  float *alpha_c, float *beta_c)
{
  const float k = 2.0 / 3.0;
  *alpha_c = k * (u_c - (v_c + w_c) / 2.0);
  *beta_c = k * SQRT_3 * (v_c - w_c) / 2.0;
}

static void
clarke_inverse (float alpha_c, float beta_c,
		float *u_c, float *v_c, float *w_c)
		 
{
  *u_c = alpha_c;
  *v_c = (-alpha_c + SQRT_3 * beta_c) / 2;
  *w_c = (-alpha_c - SQRT_3 * beta_c) / 2;
}

static void
park_transform (float alpha_c, float beta_c,
		float *q_c, float *d_c)
{
  *d_c =  alpha_c * cos_angle + beta_c * sin_angle;
  *q_c = -alpha_c * sin_angle + beta_c * cos_angle;
}

static void
park_inverse (float q_c, float d_c,
	      float *alpha_c, float *beta_c)
{
  *alpha_c = d_c * cos_angle - q_c * sin_angle;
  *beta_c  = d_c * sin_angle + q_c * cos_angle;
}

/* This routine is called once per PWM cycle, right in the middle of
   each cycle, to compute the values for the next cycle.  I.e. we're
   called every 50 uS (20 KHz).  Information we have at our disposal:

   float motor_current = actual current in amps
   int   encoder_rpm() = actual RPM
   float motor_volts()
   Pos   encoder_lval() = actual position
   int   encoder_ang () = 0..1999 relative angle

*/

static volatile float torque;
static volatile Position off_by;
static int need_quench = 1;
static short old_step;

motor_pwm_update ()
{
  int abs_rpm = encoder_rpm ();
  motor_pwm_ticks ++;
  motor_pwm_position = encoder_lval ();
  motor_pwm_rpm = abs_rpm * encoder_direction;

  switch (motor_mode)
    {
    /* -------------------------------------------------- */
    case MOTOR_ESTOP:
      {
	led_color (led_red);
	cpwm_torque (0);
	target_position = encoder_lval ();
	off_by = 0;
	torque = 0.0;
	MTU.TOERA.BYTE = 0x00;
	PORT7.DR.BYTE = UL | VL | WL;
	break;
      }
    /* -------------------------------------------------- */
    case MOTOR_STOP:
      {
	led_color (led_red);
	cpwm_torque (0);
	target_position = encoder_lval ();
	off_by = 0;
	torque = 0.0;
	if (encoder_rpm() > 500)
	  {
	    MTU.TOERA.BYTE = 0x00;
	    PORT7.DR.BYTE = 0;
	  }
	else
	  motor_mode = MOTOR_ESTOP;
	break;
      }
    /* -------------------------------------------------- */
    case MOTOR_TRACK:
      {
	int count = (short)STEP_COUNT;
	int step = (short)((short)count - (short)old_step);
	if (step)
	  {
	    old_step = count;
	    target_position += step * config_steps.position_change_per_step;
	  }
	/* fall through */
      }
    /* -------------------------------------------------- */
    case MOTOR_TRAVEL:
      {
	int abs_off;
	static Position last_target = 0;
	float volts;
	int known_lag;

	if (target_position != last_target)
	  {
	    last_target = target_position;
	  }

	if (SD_ESTOP_IN == 0)
	  {
	    cpwm_levels (0, 0, 0);
	    led_color (led_red);
	    off_by = 0;
	    torque = 0;
	    desired_rpm = 0;
	    desired_current = 0;
	    return;
	  }

#if 1
	if (motor_mode == MOTOR_TRACK
	    && tty2_can_accept (10))
	  {
	    tty2_put_marker ();
	    tty2_putw_encoded (motor_pwm_position);
	    tty2_putw_encoded (target_position);
	  }
#endif

#if 0
	if (target_position - motor_pwm_position < -2 * config_steps.position_change_per_step
	    || target_position - motor_pwm_position > 2 * config_steps.position_change_per_step)
	  pos_pid.i_enabled = 0;
	else
	  pos_pid.i_enabled = 1;
#endif

	desired_rpm = pid_iter (target_position,
				motor_pwm_position,
				motor_pwm_rpm,
				&pos_pid);

      calib_rpm:
	if (PLIMIT_HIT() && desired_rpm > 0)
	  desired_rpm = 0;
	if (NLIMIT_HIT() && desired_rpm < 0)
	  desired_rpm = 0;

	desired_current = pid_iter (desired_rpm,
				    motor_pwm_rpm,
				    desired_rpm,
				    &rpm_pid);

	if (desired_current < -config_motor_calib.peak_amps)
	  {
	    desired_current = -config_motor_calib.peak_amps;
	    rpm_pid.i_enabled = 0;
	    if (rpm_pid.i_val > 0)
	      rpm_pid.i_val = 0;
	  }
	else if (desired_current > config_motor_calib.peak_amps)
	  {
	    desired_current = config_motor_calib.peak_amps;
	    rpm_pid.i_enabled = 0;
	    if (rpm_pid.i_val < 0)
	      rpm_pid.i_val = 0;
	  }
	else
	  rpm_pid.i_enabled = 1;

#if 1
	  if (motor_mode == MOTOR_CALIBRATE_RPM_LOOP
	      && tty2_can_accept (13))
	    {
	      tty2_put_marker ();
	      tty2_putw_encoded (desired_rpm);
	      tty2_putw_encoded (motor_pwm_rpm);
	      tty2_putw_encoded (desired_current * 100);
	      tty2_putw_encoded (rpm_pid.i_val * 100);
	    }
#endif

      calib_current:
	{
	  float alpha_curr, beta_curr;
	  float q_curr, d_curr;
	  float new_q, new_d;
	  float new_alpha, new_beta;
	  float new_u, new_v, new_w;
	  int encoder_angle = encoder_ang ();


	  if (motor_current > 8.0)
	    {
	      cpwm_levels (0, 0, 0);
	      alpha_curr = beta_curr = 0;
	      q_curr = d_curr = 0;
	      new_q = new_d = 0;
	      new_alpha = new_beta = 0;
	      new_u = new_v = new_w = 0;
	    }
	  else

	    {
	      /* Given "desired_current" is the input to the standard
		 clarke-parke FOC system... */

	      /* compute our actual dq vector */
	      clarke_transform (u_curr, v_curr, w_curr, &alpha_curr, &beta_curr);
	      park_transform (alpha_curr, beta_curr, &q_curr, &d_curr);

	      new_q = pid_iter (desired_current, q_curr, desired_current, &q_pid);
	      new_d = pid_iter (0, d_curr, 0, &d_pid);

	      park_inverse (new_q, new_d, &new_alpha, &new_beta);
	      clarke_inverse (new_alpha, new_beta, &new_u, &new_v, &new_w);

	      cpwm_levels (new_u * 32767, new_v * 32767, new_w * 32767);
	    }

#if 1
	  if (motor_mode == MOTOR_TRAVEL
	      && tty2_can_accept (128))
	    {
	      tty2_put_marker ();
	      tty2_putw_encoded (target_position - calib_target_position);
	      tty2_putw_encoded (motor_pwm_position - calib_target_position);
	      tty2_putw_encoded (pos_pid.i_val * 100);
	      tty2_putw_encoded (desired_rpm);
	      tty2_putw_encoded (motor_pwm_rpm);
	      tty2_putw_encoded (rpm_pid.i_val * 100);
	      tty2_putw_encoded (desired_current * 100);
	    }
#endif
#if 1
	  if (motor_mode == MOTOR_CALIBRATE_CURRENT_LOOP
	      && tty2_can_accept (128))
	    {
	      tty2_put_marker ();
	      tty2_putc_encoded (desired_current * 10);
	      tty2_putc_encoded (u_curr * 100);
	      tty2_putc_encoded (v_curr * 100);
	      tty2_putc_encoded (w_curr * 100);
	    }
#endif
#if 0
	  if (motor_mode == MOTOR_CALIBRATE_CURRENT_LOOP
	      && tty2_can_accept (128))
	    {
	      tty2_put_marker ();
	      tty2_putc_encoded (desired_current * 10);
	      tty2_putw_encoded (u_curr * 100);
	      tty2_putw_encoded (v_curr * 100);
	      tty2_putw_encoded (w_curr * 100);
	      tty2_putw_encoded (alpha_curr * 100);
	      tty2_putw_encoded (beta_curr * 100);
	      tty2_putc_encoded (q_curr * 10);
	      tty2_putc_encoded (d_curr * 10);
	      tty2_putc_encoded (new_q * 100);
	      tty2_putc_encoded (new_d * 100);
	      tty2_putc_encoded (new_alpha * 100);
	      tty2_putc_encoded (new_beta * 100);
	      tty2_putc_encoded (new_u * 100);
	      tty2_putc_encoded (new_v * 100);
	      tty2_putc_encoded (new_w * 100);
	      tty2_putc_encoded (sin_angle * 100);
	      tty2_putc_encoded (cos_angle * 100);
	    }
#endif

	}

	if (desired_rpm > 50 || desired_rpm < -50)
	  led_color (led_green);
	else
	  led_off ();
	break;
      }

    /* -------------------------------------------------- */
    case MOTOR_RPM:
      {
	int d;
	rpm_delta += rpm_pos_per_loop;
	d = rpm_delta;
	target_position += d;
	rpm_delta -= d;

	off_by = target_position - encoder_lval ();
	torque = off_by / 400.0;
	if (torque < -1) torque = -1;
	if (torque > 1) torque = 1;
	cpwm_torque (torque * 32760);
	break;
      }

    /* -------------------------------------------------- */
    case MOTOR_TORQUE:
      {
	if (target_torque == 0)
	  motor_stop();
	else
	  {
	    cpwm_torque (target_torque);
	  }
	break;
      }
    /* -------------------------------------------------- */
    case MOTOR_CALIBRATE:
      {
	off_by = 0;
	torque = 0.0;
	break;
      }
    /* -------------------------------------------------- */
    case MOTOR_CALIBRATE_RPM_LOOP:
      {
	desired_rpm = calib_desired_rpm;
	off_by = 500;

	goto calib_rpm;
      }
    /* -------------------------------------------------- */
    case MOTOR_CALIBRATE_CURRENT_LOOP:
      {
	desired_current = calib_desired_current;
	off_by = 500;

	goto calib_current;
      }
    }
}
void
motor_estop ()
{
  cpwm_torque (0);
  motor_mode = MOTOR_ESTOP;
  MTU.TOERA.BYTE = 0x00;
  PORT7.DR.BYTE = UL | VL | WL;
}

void
motor_stop ()
{
  motor_mode = MOTOR_STOP;
}

void
motor_travel (Position new_position)
{
  bldc_run ();
  motor_mode = MOTOR_TRAVEL;
  target_position = new_position;
}

void
motor_track ()
{
  bldc_run ();
  motor_mode = MOTOR_TRACK;
  target_position = encoder_lval ();
  old_step = STEP_COUNT;
}

void
motor_step (Position delta)
{
  target_position += delta;
}

void
motor_rpm (int new_rpm)
{
  bldc_run ();
  motor_mode = MOTOR_RPM;
  target_position = encoder_lval();
  target_rpm = new_rpm;

  rpm_pos_per_loop = new_rpm / 60.0 * 4000.0 / 20000.0;
  rpm_delta = 0.0;

#if 0
  printf("\033[9;1HRPM set to %d, per %f   \n", new_rpm, rpm_pos_per_loop);
#endif
}

void
motor_torque (int new_torque)
{
  motor_mode = MOTOR_TORQUE;

  if (new_torque < -32767) new_torque = -32767;
  if (new_torque > 32767) new_torque = 32767;

  target_torque = new_torque;
  bldc_run ();
}

void
motor_calibrate ()
{
  motor_mode = MOTOR_CALIBRATE;
}

void
motor_calibrate_rpm_loop ()
{
  calib_desired_rpm = 0;
  pid_reset (&rpm_pid);
  pid_reset (&q_pid);
  pid_reset (&d_pid);
  bldc_run ();
  motor_mode = MOTOR_CALIBRATE_RPM_LOOP;
}

void
motor_calibrate_current_loop ()
{
  calib_desired_current = 0;
  pid_reset (&q_pid);
  pid_reset (&d_pid);
  bldc_run ();
  motor_mode = MOTOR_CALIBRATE_CURRENT_LOOP;
}

void
motor_status ()
{
  switch (motor_mode)
    {
    case MOTOR_ESTOP:
      printf("ESTOP");
      break;
    case MOTOR_STOP:
      printf("STOP");
      break;
    case MOTOR_TRACK:
    case MOTOR_TRAVEL:
      printf("TRAVEL rpm %f cur %f", desired_rpm, desired_current);
      break;
    case MOTOR_RPM:
      printf("RPM %d %f %f", target_rpm, rpm_pos_per_loop, rpm_delta);
      break;
    case MOTOR_CALIBRATE:
      printf("CALIBRATE");
      break;
    }
  printf("\033[K\npos %lld tpos %lld off %lld torq %f\033[K\n", encoder_lval(), target_position, off_by, torque);
}

motor_init ()
{
  pid_init (&pos_pid, &config_motor_calib.pos_pid);
  pid_init (&rpm_pid, &config_motor_calib.rpm_pid);
  pid_init (&q_pid, &config_motor_calib.current_q_pid);
  pid_init (&d_pid, &config_motor_calib.current_d_pid);

#if 0
  {
    float u, v, w, a, b;
    u = 1.5;
    v = -1.4;
    w = -0.2;
    clarke_transform (u, v, w, &a, &b);
    printf("%g %g %g -> %g %g\n", u, v, w, a, b);
    while (1);
  }
#endif
}
